"""**Utilities to support other code in the MahaUtils package**

The :py:mod:`mahautils.utils` module provides utility functions and other
general-purpose code applicable to a variety of MahaUtils modules.
"""

from .arrays import to_np_1D_array
from .capture_printing import CaptureStderr
from .dictionary import Dictionary
