"""Packages to support research work at the Maha Fluid Power Research Center

This package contains a number of utilities that are useful for performing
research work and interacting with the Maha Multics software.
"""


# PROGRAM VERSION ------------------------------------------------------------
__version__ = '0.2.0'


# PACKAGE MODULES ------------------------------------------------------------
from . import multics
from . import shapes
from . import utils
