"""Code to set up configuration panel of the simulation results viewer.
"""

from .utils_generate_table import generate_file_table_body
from .main_panel import simviewer_config_panel
from .settings_general import render_general_settings
from .settings_x import render_x_settings
from .settings_y import render_y_settings
